/*  Dynkin project programs
 *  Copyright (C) 1998 Anders S. Buch (abuch@mi.aau.dk)
 *  See the file LICENSE for license information.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>

#include "alloc.h"
#include "list.h"
#include "matrix.h"
#include "affine.h"
#include "output.h"
#include "calc.h"
#include "tree.h"
#include "cmdline.h"
#include "figlib.h"


int xrange;
int yrange;
int top, bottom;
int *left, *right;
int node_y_skip;
int labelx, labely, labelctr;

int mode = PTM_WEYL;  /* how to print nodes */

treebase *tb;
affine *afn;
list **perm_tab;

#define m_max(a, b)	((a) > (b) ? (a) : (b))

void init_dist()
{
  int max_ydist;
  int max_xdist;
  int min_x_between;
  int l, d, farleft;
  int n = m_dim(afn->cartan);
  int maxlen = tb_max_length(tb);
  
  xrange = 5 * fig_xsize / 6;
  yrange = 7 * fig_ysize / 8;
  
  max_ydist = FIG_RESOLUTION;
  max_xdist = FIG_RESOLUTION;
  min_x_between = 3 * FIG_CHAR_XSZ;
  node_y_skip = FIG_CHAR_YSZ;
  
  d = max_ydist * maxlen;
  if (yrange > FIG_CHAR_YSZ + d)
    {
      top = (yrange - d) / 2;
      bottom = (yrange + d) / 2;
    }
  else
    {
      top =  (FIG_CHAR_YSZ + 1) / 2;
      bottom = yrange - (FIG_CHAR_YSZ + 1) / 2;
    }
  
  if (mode == PTM_WEYL)
    farleft = (xrange - FIG_CHAR_XSZ * maxlen) / 2 - FIG_RESOLUTION;
  else if (mode == PTM_WEIGHT)
    farleft = (xrange - FIG_CHAR_XSZ * n) / 2 - FIG_RESOLUTION;
  else
    farleft = xrange / 2 - FIG_CHAR_XSZ * n - FIG_RESOLUTION;
  
  left = weyl_malloc((tb_max_length(tb) + 1) * sizeof(int));
  right = weyl_malloc((tb_max_length(tb) + 1) * sizeof(int));
  
  for (l = 0; l <= maxlen; l++)
    if (tb_node_count(tb, l) != 1)
      {
	int ncount = tb_node_count(tb, l);
	int node_width;
	int mx;
	
	if (mode == PTM_WEYL)
	  node_width = m_max(1, l) * FIG_CHAR_XSZ;
	else if (mode == PTM_WEIGHT)
	  node_width = n * FIG_CHAR_XSZ;
	else
	  node_width = 2 * n * FIG_CHAR_XSZ;
	
	mx = m_max(max_xdist, node_width + min_x_between);
	if ((xrange - node_width) > mx * (ncount - 1))
	  {
	    left[l] = (xrange - mx * (ncount - 1)) / 2;
	    right[l] = (xrange + mx * (ncount - 1)) / 2;
	  }
	else
	  {
	    left[l] = (node_width + 1) / 2;
	    right[l] = xrange - (node_width + 1) / 2;
	  }
	
	d = left[l] - node_width / 2;
	if (d < farleft)
	  farleft = d;
      }
  
  if (maxlen == 0)
    {
      labely = yrange / 2 - 3 * FIG_CHAR_YSZ;
      labelx = xrange / 2;
      labelctr = 1;
    }
  else
    {
      labelx = farleft;
      labely = top + FIG_CHAR_YSZ / 2;
      labelctr = 0;
    }
}


void node_pos(node *nd, int *x, int *y)
{
  int num = n_number(nd);
  int len = n_length(nd);
  int ncount = tb_node_count(tb, len);
  int maxlen = tb_max_length(tb);
  list *perm = perm_tab[len];
  
  if (perm != NULL && num < l_length(perm))
    num = l_intelem(perm, num);
  
  if (ncount == 1)
    *x = xrange / 2;
  else
    *x = (left[len] * (ncount - 1 - num)
	  + right[len] * num) / (ncount - 1);
  
  if (maxlen == 0)
    *y = yrange / 2;
  else
    *y = (top * len + bottom * (maxlen - len)) / maxlen;
}


void plot_node(node *nd)
{
  char elem_str[1000];
  int x, y, i, n = m_dim(afn->cartan);
  vector *a = n_alcove(nd);
  
  node_pos(nd, &x, &y);
  
  if (mode == PTM_WEIGHT)
    {
      for (i = 0; i < n; i++)
	{
	  int x = v_entry(a, i);
	  if (x >= 11)
	    elem_str[i] = 'A' + x - 11;
	  else if (x >= 1)
	    elem_str[i] = '0' + x - 1;
	  else
	    elem_str[i] = 'a' - x;
	}
      elem_str[n] = '\0';
    }
  else if (mode == PTM_COORD)
    {
      for (i = 0; i < n; i++)
	{
	  int x = v_entry(a, i);
	  elem_str[2 * i] = (x == 0) ? ' ' :
	    ((x < 0) ? '-' : '+');
	  x = abs(x);
	  elem_str[2 * i + 1] = (x < 10) ? '0' + x : 'A' + x - 10;
	}
      elem_str[2 * n] = '\0';
    }
  else if (n_length(nd) > 0)
    {
      list *exp = l_new();
      alcove_to_exp(exp, a, afn);
      for (i = 0; i < l_length(exp); i++)
	elem_str[i] = '0' + l_intelem(exp, i);
      elem_str[l_length(exp)] = '\0';
      l_free(exp);
    }
  else
    {
      elem_str[0] = 'e';
      elem_str[1] = '\0';
    }
  
  fig_text(x, y, elem_str, 1);
}

void connect_nodes(node *nd1, node *nd2)
{
  int x1, y1, x2, y2;
  node_pos(nd1, &x1, &y1);
  node_pos(nd2, &x2, &y2);
  fig_line(x1, y1 + node_y_skip, x2, y2 - node_y_skip);
}

#if 0
void draw_box()
{
  fig_line(0, 0, xrange, 0);
  fig_line(xrange, 0, xrange, yrange);
  fig_line(xrange, yrange, 0, yrange);
  fig_line(0, yrange, 0, 0);
}
#endif


int num, len;
list *perm;

void perm_error()
{
  cmd_error("Error in permutation file.\n");
}

void finish_perm()
{
  if (perm == NULL)
    return;
  if (perm_tab[len] != NULL)
    perm_error();
  perm_tab[len] = perm;
  perm = NULL;
}

void put_last()
{
  if (num < 0)
    return;
  if (perm == NULL)
    perm_error();
  l_appendint(perm, num);
}

#define TOK_COLON -2
int hascolon = 0;

int get_tok(FILE *prm)
{
  int c, tok;
  
  if (hascolon)
    {
      hascolon = 0;
      return TOK_COLON;
    }
  
  while (isspace(c = getc(prm)))
    ;
  if (c == EOF)
    return EOF;
  
  tok = 0;
  while (isdigit(c))
    {
      tok = 10 * tok + c - '0';
      c = getc(prm);
    }
  
  if (c == ':')
    hascolon = 1;
  else if (c != EOF && ! isspace(c))
    perm_error();
  
  return tok;
}

void read_perm_file(char *perm_nm)
{
  FILE *prm;
  int tok;
  
  prm = fopen(perm_nm, "r");
  if (prm == NULL)
    cmd_error("Can't open `%s' for reading.\n", perm_nm);
  
  num = -1;
  perm = NULL;
  
  while ((tok = get_tok(prm)) != EOF)
    {
      if (tok == TOK_COLON)
	{
	  finish_perm();
	  if (num < 0 || num > tb_max_length(tb))
	    perm_error();
	  perm = l_new();
	  len = num;
	  num = -1;
	}
      else
	{
	  put_last();
	  num = tok;
	}
    }
  put_last();
  finish_perm();
  
  fclose(prm);
}

void print_usage(char *s)
{
  fprintf(stderr, "Usage: %s [-drwc] [-z pagesize] [-p permfile] [-o output] rsys-p wge\n", s);
  exit(1);
}

int main(int ac, char **av)
{
  int n, i, l;
  node *nd;
  list *exp;
  FILE *output;
  char typ[20];
  char outbuf[500], *output_nm = NULL;
  char permbuf[500], *perm_nm = NULL;
  int filter = TB_ALL;
  int c;
  
  cmd_init(ac, av);
  fig_set_page(NULL);
  
  while ((c = getopt(ac, av, "drwcz:p:o:")) != EOF)
    switch (c)
      {
      case 'd':
	filter = TB_DOMINANT;
	break;
	
      case 'r':
	filter = TB_RESTRICTED;
	break;
	
      case 'w':
	mode = PTM_WEIGHT;
	break;
	
      case 'c':
	mode = PTM_COORD;
	break;
	
      case 'z':
	if (fig_set_page(optarg) != 0)
	  cmd_error("Unknown page size `%s'\n", optarg);
	break;
	
      case 'o':
	strncpy(outbuf, optarg, sizeof(outbuf));
	output_nm = outbuf;
	break;
	
      case 'p':
	strncpy(permbuf, optarg, sizeof(permbuf));
	perm_nm = permbuf;
	break;
	
      default:
	print_usage(av[0]);
      }
  
  afn = cmd_rsys();
  exp = cmd_wge(afn);
  n = afn_dim(afn);
  
  tb = build_tree(exp, filter, afn);
  if (tb_node_count(tb, 0) == 0)
    cmd_error("Graph is empty.\n");
  
  perm_tab = weyl_calloc(tb_max_length(tb) + 1, sizeof(list *));
  if (perm_nm)
    read_perm_file(perm_nm);
  
  output = stdout;
  if (output_nm)
    {
      output = fopen(output_nm, "w");
      if (output == NULL)
	cmd_error("Can't open `%s' for writing.\n");
    }
  
  fig_init(output);
  if (mode == PTM_WEYL)
    sprintf(typ, "%c%d", afn->root_system_type, n);
  else
    sprintf(typ, "%c%d-%d", afn->root_system_type, n, afn->p);
  
  init_dist();
  fig_text(labelx, labely, typ, labelctr);
  /* draw_box(); */
  
  for (l = tb_max_length(tb); l >= 0; l--)
    for (nd = tb_node_list(tb, l); nd != NULL; nd = n_next(nd))
      {
	plot_node(nd);
	
	for (i = 0; i < n_nbelow(nd); i++)
	  connect_nodes(nd, n_below(nd, i));
      }
  
  if (output_nm)
    fclose(output);
  
  for (i = 0; i <= tb_max_length(tb); i++)
    if (perm_tab[i])
      l_free(perm_tab[i]);
  weyl_free(perm_tab);
  tb_free(tb);
  l_free(exp);
  weyl_free(left);
  weyl_free(right);
  memory_report;
  
  exit(0);
}
