/*  Dynkin project programs
 *  Copyright (C) 1998 Anders S. Buch (abuch@mi.aau.dk)
 *  See the file LICENSE for license information.
 */

#include <stdio.h>

#if 0
#define BF_DEBUG
#endif

#include "alloc.h"
#include "file.h"
#include "list.h"
#include "matrix.h"

char fn_digit[] = ("0123456789abcdefghijklmnopqrstuvwxyz"
		   "ABCDEFGHIJKLMNOPQRSTUVWXYZ");

static char binfile_head[] = "#BF10^";
#define HEAD_SIZE	6

#if 0
#define DEBUG_WRITE_EXCEPTIONS
#endif

int unsigned_bits(unsigned maxval)
{
  int i;
  i = 0;
  while (maxval)
    {
      i++;
      maxval >>= 1;
    }
  return i;
}

binfile *bf_fopenwrite(FILE *stream, int dflt_bits, int flags)
{
  binfile *bf;
  
  if (stream == NULL)
    return NULL;
  
  bf = (binfile *) weyl_malloc(sizeof(binfile));
  bf->flags = BF_WRITE | (flags & BF_DONT_CLOSE);
  bf->stream = stream;
  bf->dflt_bits = 8;
  bf->bits_used = 0;
  bf->buf = 0;
  
  fwrite(binfile_head, HEAD_SIZE, 1, bf->stream);
  
  if (dflt_bits != 8)
    put_dflt_bits(bf, dflt_bits);
  return bf;
}

binfile *bf_openwrite(char *fname, int dflt_bits)
{
  FILE *stream = fopen(fname, "w");
  return bf_fopenwrite(stream, dflt_bits, 0);
}

binfile *bf_fopenread(FILE *stream, int flags)
{
  char head[HEAD_SIZE];
  int i;
  binfile *bf;
  
  if (stream == NULL)
    return NULL;
  
  bf = (binfile *) weyl_malloc(sizeof(binfile));
  bf->flags = BF_READ | (flags & BF_DONT_CLOSE);
  bf->stream = stream;
  bf->dflt_bits = 8;
  bf->bits_used = 0;
  bf->buf = 0;
  
  fread(head, HEAD_SIZE, 1, bf->stream);
  for (i = 0; i < HEAD_SIZE && binfile_head[i] == head[i]; i++)
    ;
  if (i < HEAD_SIZE)
    {
      fclose(bf->stream);
      weyl_free(bf);
      return NULL;
    }
  
  return bf;
}

binfile *bf_openread(char *fname)
{
  FILE *stream = fopen(fname, "r");
  return bf_fopenread(stream, 0);
}

void put_code(binfile *bf, unsigned code)
{
  put_bits(bf, ~0, bf->dflt_bits);
  put_bits(bf, code, 8);
}

void put_dflt_bits(binfile *bf, int bits)
{
  put_code(bf, SET_DEFAULT_BITS);
  put_bits(bf, bits, 8);
  bf->dflt_bits = bits;
}

static void flush_buffer(binfile *bf)
{
  unsigned y;
  if (bf->bits_used == 0)
    return;
  y = bf->buf;
  putc(y & 255, bf->stream);
  putc((y >> 8) & 255, bf->stream);
  putc((y >> 16) & 255, bf->stream);
  putc(y >> 24, bf->stream);
  bf->bits_used = 0;
  bf->buf = 0;
}

void bf_flush(binfile *bf)
{
  put_code(bf, FLUSH_FILE);
  flush_buffer(bf);
  fflush(bf->stream);
}

void bf_close(binfile *bf)
{
  if (bf->flags & BF_WRITE)
    {
      put_code(bf, END_OF_FILE);
      flush_buffer(bf);
    }
  if ((bf_flags(bf) & BF_DONT_CLOSE) == 0)
    fclose(bf->stream);
  weyl_free(bf);
}

void put_bits(binfile *bf, unsigned x, int nbits)
{
  int bu;
  unsigned mask = ~(~0 << nbits);
  unsigned xx = x & mask;
  bf->buf |= (xx << bf->bits_used);
  bu = bf->bits_used + nbits;
  if (bu < 32)
    bf->bits_used = bu;
  else
    {
      unsigned y = bf->buf;
      putc(y & 255, bf->stream);
      putc((y >> 8) & 255, bf->stream);
      putc((y >> 16) & 255, bf->stream);
      putc(y >> 24, bf->stream);
      bu -= 32;
      bf->bits_used = bu;
      bf->buf = xx >> (nbits - bu);
    }
}

unsigned get_bits(binfile *bf, int nbits)
{
  int bu = bf->bits_used;
  unsigned x;
  if (nbits <= bu)
    {
      x = bf->buf & ~(~0 << nbits);
      bf->buf >>= nbits;
      bf->bits_used = bu - nbits;
    }
  else
    {
      unsigned y;
      int d = nbits - bu;
      x = bf->buf;
      y = getc(bf->stream);
      y |= getc(bf->stream) << 8;
      y |= getc(bf->stream) << 16;
      y |= getc(bf->stream) << 24;
      x |= ((y & ~(~0 << d)) << bu);
      bf->buf = y >> d;
      bf->bits_used = 32 - d;
      if (feof(bf->stream))
	bf_flags(bf) |= BF_EOF;
    }
  if (feof(bf->stream))
    bf_flags(bf) |= BF_EOF;
#ifdef BF_DEBUG
  printf("BF %08x read %d bits: %x (%d)\n", (int) bf, nbits, x, x);
#endif
  return x;
}

void put_uint(binfile *bf, unsigned i)
{
  int bits = bf->dflt_bits;
  if (((i + 1) >> bits) != 0)
    {
      /* exceptional case */
      unsigned b, t;
#ifdef DEBUG_WRITE_EXCEPTIONS
      puts("write exception");
#endif
      put_bits(bf, ~0, bits);
      for (b = 1, t = i >> 8; t; b++)
	t = t >> 8;
      put_bits(bf, EXCPT_8BIT_ARG + b - 1, 8);
      put_bits(bf, i, 8 * b);
    }
  else
    {
      /* normal case */
      put_bits(bf, i, bits);
    }
}

void put_int(binfile *bf, int i)
{
  int bits = bf->dflt_bits;
  unsigned u = abs(i);
  if ((u >> (bits - 1)) != 0)
    {
      /* exceptional case */
      unsigned b, t;
#ifdef DEBUG_WRITE_EXCEPTIONS
      puts("write exception");
#endif
      put_bits(bf, ~0, bits);
      if (i < 0)
	u--;
      for (b = 1, t = u >> 7; t; b++)
	t = t >> 8;
      put_bits(bf, EXCPT_8BIT_ARG + b - 1, 8);
      put_bits(bf, i + (1 << (8 * b - 1)), 8 * b);
    }
  else
    {
      /* normal case */
      put_bits(bf, i - 1 + (1 << (bits - 1)), bits);
    }
}

unsigned get_uint(binfile *bf)
{
  while (! bf_eof(bf))
    {
      unsigned x = get_bits(bf, bf->dflt_bits);
      if (x == ~(~0 << bf->dflt_bits))
	{
	  unsigned tag = get_bits(bf, 8); /* puts("exception"); */
	  switch (tag)
	    {
	    case EXCPT_8BIT_ARG:
#ifdef BF_DEBUG
	      printf("BF read %08x 8BIT_ARG\n", (int) bf);
#endif
	      return get_bits(bf, 8);
	    case EXCPT_16BIT_ARG:
#ifdef BF_DEBUG
	      printf("BF read %08x 16BIT_ARG\n", (int) bf);
#endif
	      return get_bits(bf, 16);
	    case EXCPT_24BIT_ARG:
#ifdef BF_DEBUG
	      printf("BF read %08x 24BIT_ARG\n", (int) bf);
#endif
	      return get_bits(bf, 24);
	    case EXCPT_32BIT_ARG:
#ifdef BF_DEBUG
	      printf("BF read %08x 32BIT_ARG\n", (int) bf);
#endif
	      return get_bits(bf, 32);
	    case SET_DEFAULT_BITS:
	      x = get_bits(bf, 8);
#ifdef BF_DEBUG
	      printf("BF read %08x DEFAULT_BITS %d\n", (int) bf, x);
#endif
	      bf->dflt_bits = x;
	      break;
	    case FLUSH_FILE:
#ifdef BF_DEBUG
	      printf("BF read %08x FLUSH\n", (int) bf);
#endif
	      bf->bits_used = 0;
	      bf_flags(bf) |= BF_FLUSH;
	      break;
	    case END_OF_FILE:
#ifdef BF_DEBUG
	      printf("BF read %08x EOF\n", (int) bf);
#endif
	      bf_flags(bf) |= BF_EOF | BF_FLUSH;
	      break;
	    }
	}
      else
	return x;
    }
  return 0;
}

int get_int(binfile *bf)
{
  while (! bf_eof(bf))
    {
      unsigned x = get_bits(bf, bf->dflt_bits);
      if (x == ~(~0 << bf->dflt_bits))
	{
	  unsigned tag = get_bits(bf, 8);
	  switch (tag)
	    {
	    case EXCPT_8BIT_ARG:
	      return get_bits(bf, 8) - 128;
	    case EXCPT_16BIT_ARG:
	      return get_bits(bf, 16) - (1 << 15);
	    case EXCPT_24BIT_ARG:
	      return get_bits(bf, 24) - (1 << 23);
	    case EXCPT_32BIT_ARG:
	      return get_bits(bf, 32) - (1 << 31);
	    case SET_DEFAULT_BITS:
	      x = get_bits(bf, 8);
	      bf->dflt_bits = x;
	      break;
	    case FLUSH_FILE:
	      bf->bits_used = 0;
	      bf_flags(bf) |= BF_FLUSH;
	      break;
	    case END_OF_FILE:
	      bf_flags(bf) |= BF_EOF | BF_FLUSH;
	      break;
	    }
	}
      else
	return x + 1 - (1 << (bf->dflt_bits - 1));
    }
  return 0;
}

void put_vec(binfile *bf, vector *v)
{
  int i;
  for (i = 0; i < v_length(v); i++)
    put_int(bf, v_entry(v, i));
}

void get_vec(binfile *bf, vector *v)
{
  int i;
  for (i = 0; i < v_length(v) && ! bf_eof(bf); i++)
    v_entry(v, i) = get_int(bf);
}

void put_intlist(binfile *bf, list *lst)
{
  int i;
  put_uint(bf, l_length(lst));
  for (i = 0; i < l_length(lst); i++)
    put_int(bf, l_intelem(lst, i));
}

void get_intlist(binfile *bf, list *lst)
{
  int i;
  int ln = get_uint(bf);
  l_reset(lst);
  for (i = 0; i < ln && ! bf_eof(bf); i++)
    l_appendint(lst, get_int(bf));
}

void put_uvec(binfile *bf, vector *v)
{
  int i;
  for (i = 0; i < v_length(v); i++)
    put_uint(bf, v_entry(v, i));
}

void get_uvec(binfile *bf, vector *v)
{
  int i;
  for (i = 0; i < v_length(v) && ! bf_eof(bf); i++)
    v_entry(v, i) = get_uint(bf);
}

void put_ulist(binfile *bf, list *lst)
{
  int i;
  put_uint(bf, l_length(lst));
  for (i = 0; i < l_length(lst); i++)
    put_uint(bf, l_intelem(lst, i));
}

void get_ulist(binfile *bf, list *lst)
{
  int i;
  int ln = get_uint(bf);
  l_reset(lst);
  for (i = 0; i < ln && ! bf_eof(bf); i++)
    l_appendint(lst, get_uint(bf));
}

void put_string(binfile *bf, char *s)
{
  int ln = strlen(s);
  int i;
  put_uint(bf, ln);
  for (i = 0; i < ln; i++)
    put_bits(bf, (unsigned char) s[i], 8);
}

void get_string(binfile *bf, char *s)
{
  int ln = get_uint(bf);
  int i;
  for (i = 0; i < ln && ! bf_eof(bf); i++)
    s[i] = get_bits(bf, 8);
  s[ln] = '\0';
}
