/*  Dynkin project programs
 *  Copyright (C) 1998 Anders S. Buch (abuch@mi.aau.dk)
 *  See the file LICENSE for license information.
 */

#include <stdio.h>

#include "list.h"
#include "alloc.h"
#include "matrix.h"
#include "file.h"
#include "output.h"
#include "rank.h"
#include "cmdline.h"


void print_usage(char *s)
{
  fprintf(stderr, "Usage: %s file.mat ... file.mat -o new.mat\n", s);
  exit(1);
}


#define MAX_ARGS	200

int main(int ac, char **av)
{
  char *newname = NULL;
  binfile *new;
  list *names, *files, *entry;
  char sys[255];
  int i, j, n, p, dim, current_entry;
  vector *lambda, *mu, *v;
  list *buf;
  
  cmd_init(ac, av);
  
  names = l_new();
  files = l_new();
  entry = l_new();
  
  for (i = 1; i < ac; i++)
    {
      if (strcmp(av[i], "-o") != 0)
	l_appendto(names, av[i]);
      else if (newname != NULL)
	print_usage(av[0]);
      else
	newname = av[++i];
    }
  if (l_length(names) == 0 || newname == NULL)
    print_usage(av[0]);
  
  for (i = 0; i < l_length(names); i++)
    if (strcmp(newname, l_element(names, i)) == 0)
      cmd_error("Output file must be different from input files.\n");
  
  n = p = dim = 0;
  lambda = mu = v = NULL;
  for (i = 0; i < l_length(names); i++)
    {
      char s[255];
      char *name = l_element(names, i);
      binfile *bf = bf_openread(name);
      l_appendto(files, bf);
      if (bf == NULL)
	cmd_error("Bad matrix file `%s'\n", name);
      
      if (i == 0)
	{
	  get_string(bf, sys);
	  n = atoi(sys + 1);
	  p = get_uint(bf);
	  lambda = v_new(n);
	  mu = v_new(n);
	  v = v_new(n);
	  get_uvec(bf, lambda);
	  get_uvec(bf, mu);
	  dim = get_uint(bf);
	}
      else
	{
	  int d1, d2, d3, d4;
	  get_string(bf, s);
	  d4 = p - get_uint(bf);
	  get_uvec(bf, v);
	  d1 = v_compare(lambda, v);
	  get_uvec(bf, v);
	  d2 = v_compare(mu, v);
	  d3 = dim - get_uint(bf);
	  if (strcmp(sys, s) != 0 
	      || d1 != 0 || d2 != 0 || d3 != 0 || d4 != 0)
	    cmd_error("Input files inconsistent.\n");
	}
      l_appendint(entry, get_uint(bf));
    }
  
  l_free(names);
  v_free(v);
  
  printf("Root system:  %s, p = %d\n", sys, p);
  printf("lambda = ");
  for (i = 0; i < n; i++)
    putchar(fn_digit[v_entry(lambda, i)]);
  printf(" ;  mu = ");
  for (i = 0; i < n; i++)
    putchar(fn_digit[v_entry(mu, i)]);
  putchar('\n');
  printf("Matrix dimension:  %d\n", dim);
  
  for (i = 1; i < l_length(entry); i++)
    for (j = 0; j < i; j ++)
      {
	if (l_intelem(entry, i) < l_intelem(entry, j))
	  {
	    int x;
	    binfile *bf = l_element(files, i);
	    l_element(files, i) = l_element(files, j);
	    l_element(files, j) = bf;
	    x = l_intelem(entry, i);
	    l_intelem(entry, i) = l_intelem(entry, j);
	    l_intelem(entry, j) = x;
	  }
      }
  current_entry = l_intelem(entry, 0);
  
  if ((new = bf_openwrite(newname, 8)) == NULL)
    cmd_error("Can't open `%s' for writing.\n");
  
  put_string(new, sys);
  put_uint(new, p);
  put_uvec(new, lambda);
  put_uvec(new, mu);
  put_uint(new, dim);
  put_uint(new, current_entry);
  i = unsigned_bits(p);
  put_dflt_bits(new, i);
  
  
  buf = l_new();
  for (i = 0; i < l_length(files); i++)
    {
      binfile *bf = l_element(files, i);
      int file_ent = l_intelem(entry, i);
      
      if (file_ent > current_entry)
	{
	  fprintf(stderr, "%s: Warning: Entry range gap; "
		  "%d files skiped.\n", av[0], l_length(files) - i);
	  i = l_length(files);
	}
      else
	while (! bf_eof(bf))
	  {
	    int x = get_uint(bf);
	    if (bf_isflushed(bf))
	      {
		for (j = current_entry - file_ent; j < l_length(buf); j++)
		  {
		    put_uint(new, l_intelem(buf, j));
		    current_entry++;
		  }
		
		file_ent += l_length(buf);
		l_reset(buf);
		
		bf_clearflush(bf);
	      }
	    l_appendint(buf, x);
	  }
      
      bf_close(bf);
      l_reset(buf);
    }

#if 0
  entries_total = dim * (dim + 1) / 2;
  percent = 1000 * entries_read / entries_total;
  printf("Entries in file:   %d  of  %d   (%d.%d%%)\n", entries_read, 
	 entries_total, percent / 10, percent % 10);
#endif
  
  l_free(buf);
  v_free(lambda);
  v_free(mu);
  l_free(files);
  l_free(entry);
  bf_close(new);
  
  memory_report;
  
  exit(0);
}
