/*  Dynkin project programs
 *  Copyright (C) 1998 Anders S. Buch (abuch@mi.aau.dk)
 *  See the file LICENSE for license information.
 */

#include <stdio.h>

#include "alloc.h"
#include "matrix.h"
#include "file.h"
#include "output.h"
#include "rank.h"
#include "cmdline.h"


void print_usage(char *s)
{
  fprintf(stderr, "Usage: %s file.mat\n", s);
  exit(1);
}


typedef struct _charmat {
  int dim;
  char e[1];
} charmat;

#define cm_dim(m)		((m)->dim)
#define cm_entry(m, i, j)	((m)->e[(i) * (m)->dim + (j)])
#define cm_set_zero(m)		\
	memset((m)->e, 0, (m)->dim * (m)->dim * sizeof(char))
#define cm_copy(d, m)		\
	memcpy((d)->e, (m)->e, (d)->dim * (d)->dim * sizeof(char))

charmat *cm_new(int dim)
{
  charmat *cm = 
    (charmat *) weyl_malloc(sizeof(charmat) + (dim * dim - 1) * sizeof(char));
  cm_dim(cm) = dim;
  return cm;
}

#ifdef __GNUC__
inline
#endif
static int mod(int a, int p)
{
  a %= p;
  return (a < 0) ? a + p : a;
}

int cm_rank_mod_p(charmat *cm, int p, vector *inv)
{
  int r, c, i, j, x;
  int n = cm_dim(cm);
  
  r = 0;
  c = 0;
  while (c < n)
    {
      i = r;
      while (i < n && cm_entry(cm, i, c) == 0)
	i++;
      if (i < n)
	{
	  if (i != r)
	    {
	      for (j = c; j < n; j++)
		{
		  int tmp = cm_entry(cm, r, j);
		  cm_entry(cm, r, j) = cm_entry(cm, i, j);
		  cm_entry(cm, i, j) = tmp;
		}
	    }
	  x = v_entry(inv, (int) cm_entry(cm, r, c));
	  for (i = r + 1; i < n; i++)
	    {
	      int t = x * cm_entry(cm, i, c);
	      for (j = c + 1; j < n; j++)
		cm_entry(cm, i, j) = 
		  mod(cm_entry(cm, i, j) - t * cm_entry(cm, r, j), p);
	    }
	  r++;
	}
      c++;
    }
  return r;
}

int main(int ac, char **av)
{
  binfile *bf;
  char s[255];
  int n, p, dim, i, j;
  charmat *M;
  vector *v, *inv;
  int first_entry;
  char *name;
  
  cmd_init(ac, av);
  name = cmd_file(1);
  
  if ((bf = bf_openread(name)) == NULL)
    cmd_error("Bad matrix file `%s'\n", name);
  
  get_string(bf, s);
  puts(s);
  n = atoi(s + 1);
  p = get_uint(bf);
  inv = calc_inverse_list(p);
  
  v = v_new(n);
  get_uvec(bf, v);
  print_vec(v, 0);
  get_uvec(bf, v);
  print_vec(v, 0);
  
  dim = get_uint(bf);
  printf("dim = %d\n", dim);
  M = cm_new(dim);
  
  first_entry = get_uint(bf);
  if (first_entry != 0)
    cmd_error("Matrix file incomplete.\n");
  
  for (i = 0; i < dim; i++)
    {
      for (j = 0; j <= i; j++)
	{
	  int x = get_uint(bf);
	  m_entry(M, i, j) = x;
	  m_entry(M, j, i) = x;
	}
    }
  
  if (bf_eof(bf))
    cmd_error("Matrix file incomplete.\n");
  
  printf("matrix read from file.\n");
  fflush(stdout);
  
  printf("rank: %d\n", cm_rank_mod_p(M, p, inv));
  
  v_free(v);
  m_free(M);
  v_free(inv);
  bf_close(bf);
  
  memory_report;
  
  exit(0);
}
