/*  Dynkin project programs
 *  Copyright (C) 1998 Anders S. Buch (abuch@mi.aau.dk)
 *  See the file LICENSE for license information.
 */

#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "alloc.h"
#include "matrix.h"

vector *v_new(int sz)
{
  vector *v = (vector *) weyl_malloc(sizeof(vector) + (sz - 1) * sizeof(int));
  v_length(v) = sz;
  return v;
}

vector *v_new_zero(int sz)
{
  vector *v = v_new(sz);
  v_set_zero(v);
  return v;
}

vector *v_new_unit(int sz, int i)
{
  vector *v = v_new(sz);
  v_set_unit(v, i);
  return v;
}

vector *v_new_copy(vector *src)
{
  vector *v = v_new(v_length(src));
  v_copy(v, src);
  return v;
}

void v_set_ones(vector *v)
{
  int i;
  for (i = 0; i < v_length(v); i++)
    v_entry(v, i) = 1;
}

void v_add(vector *dst, vector *v1, vector *v2)
{
  int i;
  for (i = 0; i < v_length(dst); i++)
    v_entry(dst, i) = v_entry(v1, i) + v_entry(v2, i);
}

void v_sub(vector *dst, vector *v1, vector *v2)
{
  int i;
  for (i = 0; i < v_length(dst); i++)
    v_entry(dst, i) = v_entry(v1, i) - v_entry(v2, i);
}

void v_scale(vector *dst, int r, vector *v)
{
  int i;
  for (i = 0; i < v_length(dst); i++)
    v_entry(dst, i) = r * v_entry(v, i);
}

void v_div(vector *dst, vector *v, int r)
{
  int i;
  for (i = 0; i < v_length(dst); i++)
    v_entry(dst, i) = v_entry(v, i) / r;
}

void v_add_multiple(vector *dst, int r, vector *v)
{
  int i;
  for (i = 0; i < v_length(dst); i++)
    v_entry(dst, i) += r * v_entry(v, i);
}

int v_dot(vector *v1, vector *v2)
{
  int i, x;
  x = 0;
  for (i = 0; i < v_length(v1); i++)
    x += v_entry(v1, i) * v_entry(v2, i);
  return x;
}

int v_sum_of_entries(vector *v)
{
  int i, sum;
  sum = 0;
  for (i = 0; i < v_length(v); i++)
    sum += v_entry(v, i);
  return sum;
}

int v_compare(vector *v1, vector *v2)
{
  int i;
  for (i = 0; i < v_length(v1) - 1 && v_entry(v1, i) == v_entry(v2, i); i++)
    ;
  return v_entry(v1, i) - v_entry(v2, i);
}

int v_lesseq(vector *v1, vector *v2)
{
  int i;
  for (i = v_length(v1); i && v_entry(v1, i - 1) <= v_entry(v2, i - 1); i--)
    ;
  return i == 0;
}

hashkey_t v_hash(vector *v)
{
  int i;
  hashkey_t g, h;

  h = v_length(v);
  for (i = 0; i < v_length(v); i++)
    {
      h = (h << 3) + v_entry(v, i);
      g = h & 0xf0000000;
      h ^= (g >> 24);
      /* h ^= g; */
    }

  return h;
}


matrix *m_new(int dim)
{
  matrix *m = 
    (matrix *) weyl_malloc(sizeof(matrix) + (dim * dim - 1) * sizeof(int));
  m_dim(m) = dim;
  return m;
}

matrix *m_new_zero(int dim)
{
  matrix *m = m_new(dim);
  m_set_zero(m);
  return m;
}

matrix *m_new_id(int dim)
{
  matrix *m = m_new(dim);
  m_set_id(m);
  return m;
}

void m_set_id(matrix *m)
{
  int i;
  m_set_zero(m);
  for (i = 0; i < m_dim(m); i++)
    m_entry(m, i, i) = 1;
}

void m_get_column(vector *v, matrix *m, int j)
{
  int i;
  for (i = 0; i < v_length(v); i++)
    v_entry(v, i) = m_entry(m, i, j);
}

void m_mult(matrix *dst, matrix *m1, matrix *m2)
{
  int i, j, k, x;
  int n = m_dim(dst);
  for (i = 0; i < n; i++)
    for (j = 0; j < n; j++)
      {
	x = 0;
	for (k = 0; k < n; k++)
	  x += m_entry(m1, i, k) * m_entry(m2, k, j);
	m_entry(dst, i, j) = x;
      }
}

void m_scale(matrix *dst, int r, matrix *m)
{
  int i, j;
  int n = m_dim(dst);
  for (i = 0; i < n; i++)
    for (j = 0; j < n; j++)
      m_entry(dst, i, j) = r * m_entry(m, i, j);
}

void m_div(matrix *dst, matrix *m, int r)
{
  int i, j;
  int n = m_dim(dst);
  for (i = 0; i < n; i++)
    for (j = 0; j < n; j++)
      m_entry(dst, i, j) = m_entry(m, i, j) / r;
}

void m_transpose(matrix *dst, matrix *m)
{
  int i, j;
  int n = m_dim(dst);
  for (i = 0; i < n; i++)
    for (j = 0; j < n; j++)
      m_entry(dst, i, j) = m_entry(m, j, i);
}


struct float_mat {
  int dim;
  int rowlength;
  double *data;
  double *rows[1];
};

#define fm_entry(fm, i, j)	(((fm)->rows[i])[j])

struct float_mat *mk_float_mat(matrix *m, int dub)
{
  int i, j;
  int n = m_dim(m);

  struct float_mat *fm = (struct float_mat *)
    weyl_malloc(sizeof(struct float_mat) + (n - 1) * sizeof(double *));

  double *data = (double *) weyl_malloc(n * n * dub * sizeof(double));
  
  fm->dim = n;
  fm->rowlength = n * dub;
  fm->data = data;
  
  for (i = 0; i < n; i++)
    fm->rows[i] = data + i * n * dub;
  
  for (i = 0; i < n; i++)
    for (j = 0; j < n; j++)
      fm_entry(fm, i, j) = m_entry(m, i, j);
  
  if (dub == 2)
    for (i = 0; i < n; i++)
      for (j = 0; j < n; j++)
	fm_entry(fm, i, j + n) = (i == j ? 1 : 0);
  
  return fm;
}

void free_float_mat(struct float_mat *fm)
{
  weyl_free(fm->data);
  weyl_free(fm);
}

#define round_off(d)	((int) (((d) >= 0) ? (d) + 0.5 : (d) - 0.5))

int fm_utreduce(struct float_mat *fm)
{
  int n = fm->dim;
  int rln = fm->rowlength;
  int i, j, k, sign;
  double det;
  
  sign = 1;
  for (j = 0; j < n; j++)
    {
      double max = 0;
      int imax = 0;
      
      for (i = j; i < n; i++)
	if (fabs(fm_entry(fm, i, j)) > max)
	  {
	    imax = i;
	    max = fabs(fm_entry(fm, i, j));
	  }
      
      if (max == 0)
	return 0;
      
      if (imax != j)
	{
	  double *tmp = fm->rows[j];
	  fm->rows[j] = fm->rows[imax];
	  fm->rows[imax] = tmp;
	  sign = -sign;
	}
      
      for (i = j + 1; i < n; i++)
	{
	  double a = fm_entry(fm, i, j) / fm_entry(fm, j, j);
	  for (k = j + 1; k < rln; k++)
	    fm_entry(fm, i, k) -= a * fm_entry(fm, j, k);
	}
    }
  
  det = sign;
  for (i = 0; i < n; i++)
    det = det * fm_entry(fm, i, i);
  
  return round_off(det);
}

void fm_back_subst(struct float_mat *fm)
{
  int n = fm->dim;
  int rln = fm->rowlength;
  int i, j, k;
  
  for (j = n - 1; j >= 0; j--)
    {
      double a = fm_entry(fm, j, j);
      for (k = n; k < rln; k++)
	fm_entry(fm, j, k) /= a;
      
      for (i = 0; i < j; i++)
	{
	  double b = fm_entry(fm, i, j);
	  for (k = n; k < rln; k++)
	    fm_entry(fm, i, k) -= b * fm_entry(fm, j, k);
	}
    }
}

int m_determinant(matrix *m)
{
  struct float_mat *fm = mk_float_mat(m, 1);
  int d = fm_utreduce(fm);
  free_float_mat(fm);
  return d;
}

void m_adjoint(matrix *dst, matrix *m, int *det)
{
  int n = m_dim(dst);
  struct float_mat *fm = mk_float_mat(m, 2);
  int d = fm_utreduce(fm);
  int i, j;
  fm_back_subst(fm);
  for (i = 0; i < n; i++)
    for (j = 0; j < n; j++)
      m_entry(dst, i, j) = round_off(d * fm_entry(fm, i, n + j));
  free_float_mat(fm);
  if (det != NULL)
    *det = d;
}


void m_multvec(vector *dst, matrix *m, vector *v)
{
  int i, k;
  int n = v_length(dst);
  for (i = 0; i < n; i++)
    {
      int x = 0;
      for (k = 0; k < n; k++)
	x += m_entry(m, i, k) * v_entry(v, k);
      v_entry(dst, i) = x;
    }
}

void m_vecmult(vector *dst, vector *v, matrix *m)
{
  int k, j;
  int n = v_length(dst);
  for (j = 0; j < n; j++)
    {
      int x = 0;
      for (k = 0; k < n; k++)
	x += v_entry(v, k) * m_entry(m, k, j);
      v_entry(dst, j) = x;
    }
}
