/*  Dynkin project programs
 *  Copyright (C) 1998 Anders S. Buch (abuch@mi.aau.dk)
 *  See the file LICENSE for license information.
 */

#include "matrix.h"
#include "redulib.h"
#include "alloc.h"
#include "list.h"
#include "efstring.h"
#include "affine.h"
#include "sconst.h"

list *build_lambda_pairs(vector *lambda, affine *afn)
{
  list *plus_roots = afn->positive_roots;
  list *root_length = afn->root_length;
  int nroots = l_length(plus_roots);
  int n = m_dim(afn->cartan);
  list *res = l_new();
  int j;
  
  vector *lr = v_new(n);
  vector *tmp = v_new(n);
  m_multvec(tmp, afn->atcartan, lambda);
  m_vecmult(lr, tmp, afn->root_dot);
  
  for (j = 0; j < nroots; j++)
    {
      vector *aj = l_element(plus_roots, j);
      l_appendint
	(res, v_dot(lr, aj) / (afn->dcartan * l_intelem(root_length, j)));
    }
  
  v_free(tmp);
  v_free(lr);
  return res;
}

static list *sums_to_eflist(vector *sum1, vector *sum2)
{
  list *efl = l_new();
  int i, j;
  for (i = v_length(sum1) - 1; i >= 0; i--)
    {
      for (j = 0; j < v_entry(sum1, i); j++)
	l_appendint(efl, efl_mkelm(E_BASE, i + 1));
    }
  for (i = 0; i < v_length(sum2); i++)
    {
      for (j = 0; j < v_entry(sum2, i); j++)
	l_appendint(efl, efl_mkelm(i + 1, F_BASE));
    }
  return efl;
}

#define REDUCE_MODULO	9699690
/*  ==  2 * 3 * 5 * 7 * 11 * 13 * 17 * 19  */

typedef struct _efl_node {
  list *efl;
  int value;
  struct _efl_node *left, *right;
} efl_node;

static list *efl_trees;
static efl_node *last_node;

static matrix *Nneg;
static matrix *sub_mat;
static matrix *pair_mat;
static list *lambda_pairs;

void init_redulib(vector *lambda, affine *afn)
{
  matrix *Npos, *add_mat;
  list *plus_roots= afn->positive_roots;
  
  efl_trees = l_new();
  
  add_mat = build_add_matrix(plus_roots, 1);
  sub_mat = build_add_matrix(plus_roots, -1);
  pair_mat = build_pair_matrix(afn);
  lambda_pairs = build_lambda_pairs(lambda, afn);
  
  Npos = m_new(l_length(plus_roots));
  Nneg = m_new(l_length(plus_roots));
  calc_sconsts(afn, add_mat, sub_mat, Npos, Nneg);
  m_free(Npos);
  m_free(add_mat);
}

void free_redulib()
{
  m_free(sub_mat);
  m_free(pair_mat);
  l_free(lambda_pairs);
  m_free(Nneg);
  l_free(efl_trees);
}


efl_node *mk_efl_node(list *efl)
{
  efl_node *node = weyl_calloc(1, sizeof(efl_node));
  node->efl = l_new();
  l_copy(node->efl, efl);
  return node;
}

efl_node *find_node(list *efl, efl_node *node)
{
  efl_node *last = NULL;
  int cmp = 0;
  
  while (node != NULL && (cmp = l_compare(efl, node->efl)) != 0)
    {
      last = node;
      node = (cmp < 0) ? node->left : node->right;
    }
  
  if (node == NULL)
    {
      last_node = mk_efl_node(efl);
      if (cmp < 0)
	last->left = last_node;
      else
	last->right = last_node;
    }
  return node;
}

efl_node *find_efl(list *efl)
{
  int ind = l_length(efl) - 1;
  efl_node *node = l_element(efl_trees, ind);
  
  if (node == NULL)
    {
      last_node = l_element(efl_trees, ind) = mk_efl_node(efl);
      return NULL;
    }
  else
    return find_node(efl, node);
}

void free_efl_node(efl_node *node)
{
  if (node == NULL)
    return;
  free_efl_node(node->left);
  free_efl_node(node->right);
  l_free(node->efl);
  weyl_free(node);
}

void free_efl_trees()
{
  int i;
  for (i = 0; i < l_length(efl_trees); i++)
    {
      free_efl_node(l_element(efl_trees, i));
      l_element(efl_trees, i) = NULL;
    }
}

int do_red(list *efl)
{
  efl_node *node;
  int typ, root;
  int i;
  
  if (l_length(efl) == 0)
    {
      l_free(efl);
      return 1;
    }
  
  if (efl_gettype(efl, 0) == F_BASE
      || efl_gettype(efl, l_length(efl) - 1) == E_BASE)
    {
      l_free(efl);
      return 0;
    }
  
  if ((node = find_efl(efl)) != NULL)
    {
      l_free(efl);
      return node->value;
    }
  node = last_node;
  
  i = l_length(efl);
  while (i > 0 && efl_gettype(efl, --i) == F_BASE)
    ;
  
  typ = efl_gettype(efl, i);
  root = efl_getroot(efl, i);
  
  if (typ == H_BASE)
    {
      if (i == l_length(efl) - 1)
	{
	  l_length(efl)--;
	  return node->value =
	    (l_intelem(lambda_pairs, root - 1) * do_red(efl)) 
	      % REDUCE_MODULO;
	}
      else
	{
	  int nxt_root = efl_getroot(efl, i + 1);
	  int pair = m_entry(pair_mat, nxt_root - 1, root - 1);
	  int tmp = 0;
	  
	  if (pair != 0)
	    {
	      list *efl2 = l_new();
	      l_copy(efl2, efl);
	      l_remove_elem(efl2, i);
	      tmp = pair * do_red(efl2);
	    }
	  
	  l_intelem(efl,i) = F_BASE + ROOT_ZERO + nxt_root;

	  efl_set(efl, i, F_BASE, nxt_root);
	  efl_set(efl, i + 1, H_BASE, root);
	  return node->value =
	    (do_red(efl) - tmp) % REDUCE_MODULO;
	}
    }
  else /* typ == E_BASE */
    {
      int nxt_root = efl_getroot(efl, i + 1);
      if (root == nxt_root)
	{
	  list *efl2 = l_new();
	  l_copy(efl2, efl);
	  efl_set(efl2, i, F_BASE, nxt_root);
	  efl_set(efl2, i + 1, E_BASE, root);
	  
	  l_remove_elem(efl, i);
	  efl_set(efl, i, H_BASE, root);
	  
	  return node->value =
	    (do_red(efl) + do_red(efl2)) % REDUCE_MODULO;
	}
      else /* root != nxt_root */
	{
	  int nval = m_entry(Nneg, root - 1, nxt_root - 1);
	  int tmp = 0;
	  
	  if (nval != 0)
	    {
	      int dif = m_entry(sub_mat, root - 1, nxt_root - 1);
	      list *efl2 = l_new();
	      l_copy(efl2, efl);
	      l_remove_elem(efl2, i);
	      if (dif > 0)
		efl_set(efl2, i, E_BASE, dif);
	      else
		efl_set(efl2, i, F_BASE, -dif);
	      tmp = nval * do_red(efl2);
	    }
	  
	  efl_set(efl, i, F_BASE, nxt_root);
	  efl_set(efl, i + 1, E_BASE, root);
	  
	  return node->value = 
	    (do_red(efl) + tmp) % REDUCE_MODULO;
	}
    }
}

int reduce(vector *sum1, vector *sum2)
{
  list *efl = sums_to_eflist(sum1, sum2);
  while (l_length(efl_trees) < l_length(efl))
    l_appendto(efl_trees, NULL);
  return do_red(efl);
}

